@@ -4,26 +4,30 @@
 """
 Tests for twisted.enterprise.pgadbapi.
 """
+import os
 
 try:
     import psycopg2
     import psycopg2.extensions
 except ImportError:
     # try psycopg2-ctypes
-    import psycopg2ct as psycopg2
+    try:
+        import psycopg2ct as psycopg2
+    except ImportError:
+        raise ImportError('no module named psycopg2 or psycopg2ct')
 
 from txpostgres import txpostgres
 
 from twisted.trial import unittest
-from twisted.internet import defer, error, main, posixbase, reactor
+from twisted.internet import defer, error, main, posixbase, reactor, task
 from twisted.python import failure
 
 simple_table_schema = "CREATE TABLE simple (x integer)"
 
-DB_NAME = "twisted_test"
-DB_HOST = "localhost"
-DB_USER = "twisted_test"
-DB_PASS = "twisted_test"
+DB_NAME = os.getenv("TXPOSTGRES_TEST_DATABASE", "twisted_test")
+DB_HOST = os.getenv("TXPOSTGRES_TEST_HOST", "localhost")
+DB_USER = os.getenv("TXPOSTGRES_TEST_USER", "twisted_test")
+DB_PASS = os.getenv("TXPOSTGRES_TEST_PASSWORD", "twisted_test")
 
 
 def getSkipForPsycopg2():
@@ -702,11 +706,14 @@ class TxPostgresQueryTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
             raise unittest.SkipTest("This test fails on versions of Twisted "
                                     "affected by Twisted bug #4539")
 
-        # skip this test if running with psycopg2ct, which does not detect that
-        # the connection has been closed until it's too late and instead of a
-        # DatabaseError gets an error because the file descriptor is -1
-        if getattr(psycopg2, '_impl', None):
-            raise unittest.SkipTest("This test fails with psycopg2ct")
+        # Check if we're running under psycopg2ct from before the patch that
+        # added correct terminated connection handling. Then only way to know
+        # that seems to be looking at the __version__ string.
+        if (getattr(psycopg2, '_impl', None) and
+            'ctypes' not in psycopg2.__version__):
+            raise unittest.SkipTest(
+                "This test fails on versions of psycopg2ct 0.3 and older, "
+                "which have a bug in terminated connection handling")
 
         def checkSuperuser(ret):
             if ret[0][0] != 'on':
@@ -878,8 +885,8 @@ class TxPostgresConnectionPoolErrorsTestCase(Psycopg2TestCase):
         class ErrorConnection(object):
             connid = 0
 
-            def __init__(self, reactor):
-                self.reactor = reactor
+            def __init__(self, *args):
+                pass
 
             def connect(self, *args, **kwargs):
                 # the third connection fails
@@ -974,34 +981,83 @@ class TxPostgresCancellationTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
                     "You need at least version 2.3.0 of psycopg2 "
                     "to use query cancellation.")
 
+        def createAdditionalConnection():
+            # Create an additional connection that will be used to check if the
+            # main one is already waiting on pg_sleep.
+            self.extra = txpostgres.Connection()
+            return self.extra.connect(user=DB_USER, password=DB_PASS,
+                                      host=DB_HOST, database=DB_NAME)
+
         d = _SimpleDBSetupMixin.setUp(self)
-        return d.addCallback(lambda _: checkCancellationSupport())
+        d.addCallback(lambda _: checkCancellationSupport())
+        return d.addCallback(lambda _: createAdditionalConnection())
+
+    def tearDown(self):
+        d = _SimpleDBSetupMixin.tearDown(self)
+        return d.addCallback(lambda _: self.extra.close())
+
+    def waitForSleep(self):
+        """
+        Return a Deferred that fires when the main connection starts executing
+        pg_sleep. This is done by polling pg_stat_statements from the extra
+        connection.
+        """
+        # XXX is this really the only way to make these tests not racy?
+        interval = 0.5
+        initial_remaining = 10
+
+        def gotResult(res, remaining):
+            if res and res[0][0]:
+                # the result of the query was True, we're done
+                return
+
+            if not remaining:
+                self.fail('main connection did not execute pg_sleep')
+
+            # main connection still not executing pg_sleep, wait
+            return task.deferLater(reactor, interval, checkActivity, remaining)
+
+        def checkActivity(remaining):
+            d = self.extra.runQuery("select current_query like '%%pg_sleep%%' "
+                                    "from pg_stat_activity where procpid = %s",
+                                    (self.conn.get_backend_pid(), ))
+            return d.addCallback(gotResult, remaining - 1)
+
+        return checkActivity(initial_remaining)
 
     def test_simpleCancellation(self):
         d = self.conn.runQuery("select pg_sleep(5)")
-        reactor.callLater(0, self.conn.cancel, d)
-        return self.failUnlessFailure(d, defer.CancelledError)
+
+        def cancelQuery(_):
+            self.conn.cancel(d)
+            return self.failUnlessFailure(d, defer.CancelledError)
+
+        waiting = self.waitForSleep()
+        return waiting.addCallback(cancelQuery)
 
     def test_directCancellation(self):
         d = self.conn.runQuery("select pg_sleep(5)")
 
-        def tryDirectCancel(d):
+        def tryDirectCancel(_):
             self.assertRaises(txpostgres._CancelInProgress, d.cancel)
-        reactor.callLater(0, tryDirectCancel, d)
+            return self.failUnlessFailure(d, defer.CancelledError)
 
-        return self.failUnlessFailure(d, defer.CancelledError)
+        waiting = self.waitForSleep()
+        return waiting.addCallback(tryDirectCancel)
 
     def test_cancelInteraction(self):
         def interaction(c):
-            def cancelAndPassthrough(ret):
-                reactor.callLater(0, self.conn.cancel, d)
-                return ret
+            def cancelQuery(_):
+                self.conn.cancel(d)
+                return d
 
             d = c.execute("insert into simple values (1)")
             d.addCallback(lambda c: c.execute("insert into simple values (2)"))
-            d.addCallback(cancelAndPassthrough)
             d.addCallback(lambda c: c.execute("select pg_sleep(5)"))
-            return d.addCallback(lambda _: "interaction done")
+            d.addCallback(lambda _: "interaction done")
+
+            waiting = self.waitForSleep()
+            return waiting.addCallback(cancelQuery)
 
         d = self.conn.runInteraction(interaction)
         d = self.failUnlessFailure(d, defer.CancelledError)
@@ -1012,13 +1068,18 @@ class TxPostgresCancellationTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
     def test_cancelMultipleQueries(self):
         d1 = self.conn.runQuery("select pg_sleep(5)")
         d2 = self.conn.runQuery("select pg_sleep(5)")
-        reactor.callLater(0, self.conn.cancel, d1)
-        reactor.callLater(0, self.conn.cancel, d2)
 
-        d1 = self.failUnlessFailure(d1, defer.CancelledError)
-        d2 = self.failUnlessFailure(d2, defer.CancelledError)
+        def cancelQueries(_):
+            self.conn.cancel(d1)
+            self.conn.cancel(d2)
 
-        return defer.gatherResults([d1, d2])
+            self.failUnlessFailure(d1, defer.CancelledError)
+            self.failUnlessFailure(d2, defer.CancelledError)
+
+            return defer.gatherResults([d1, d2])
+
+        waiting = self.waitForSleep()
+        return waiting.addCallback(cancelQueries)
 
 
 class TxPostgresNotifyObserversTestCase(Psycopg2TestCase):
@@ -1210,12 +1271,14 @@ class TxPostgresNotifyTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         """
         dl = [defer.Deferred(), defer.Deferred()]
         notifyD = defer.DeferredList(dl)
+        errorD = defer.Deferred()
 
         def observer1(notify):
             self.notifies.append(1)
             dl.pop().callback(None)
 
         def observer2(notify):
+            errorD.callback(None)
             raise RuntimeError("boom")
 
         self.conn.addNotifyObserver(observer1)
@@ -1226,9 +1289,120 @@ class TxPostgresNotifyTestCase(_SimpleDBSetupMixin, Psycopg2TestCase):
         # at some point both observer functions will get called, one of them
         # raising an exception, to make sure they still are registered and
         # executing, send another notify
+        d.addCallback(lambda _: errorD)
         d.addCallback(lambda _: self.conn.removeNotifyObserver(observer2))
         d.addCallback(lambda _: self.sendNotify())
         d.addCallback(lambda _: notifyD)
-        d.addCallback(lambda _: self.flushLoggedErrors(RuntimeError))
+        d.addCallback(lambda _: self.assertEquals(
+                len(self.flushLoggedErrors(RuntimeError)), 1))
         return d.addCallback(lambda _: self.assertEquals(
                 self.notifies, [1, 1]))
+
+    def test_observerReturnsDeferred(self):
+        """
+        If the observer function returns a Deferred, the next notify won't be
+        processed until it fires.
+        """
+        observerDs = [defer.Deferred(), defer.Deferred()]
+        notifyDs = [defer.Deferred(), defer.Deferred()]
+
+        oneObserverDone = defer.DeferredList(notifyDs, fireOnOneCallback=True)
+        allObserversDone = defer.DeferredList(notifyDs)
+
+        def fireAllObserverDs():
+            for observerD in observerDs:
+                observerD.callback(None)
+            return allObserversDone
+
+        def observer(notify, observerD, notifyD):
+            self.notifies.append(notify)
+            # don't callback notifyD synchronously, or oneObserverDone might
+            # start processing immediately
+            reactor.callLater(0, notifyD.callback, None)
+            return observerD
+
+        def makeObserver(observerD, notifyD):
+            return lambda notify: observer(notify, observerD, notifyD)
+
+        for observerD, notifyD in zip(observerDs, notifyDs):
+            self.conn.addNotifyObserver(makeObserver(observerD, notifyD))
+
+        d = self.conn.runOperation("listen txpostgres_test")
+        d.addCallback(lambda _: self.sendNotify())
+        # wait for one of the observers to finish -- there's no guarantee of
+        # order, so can't say which
+        d.addCallback(lambda _: oneObserverDone)
+        # only one observer should have been called
+        d.addCallback(lambda _: self.assertEquals(len(self.notifies), 1))
+        # fire both observer Deferreds to make sure the processing finishes
+        d.addCallback(lambda _: fireAllObserverDs())
+        # both observers should now be called
+        return d.addCallback(lambda _: self.assertEquals(
+                len(self.notifies), 2))
+
+    def test_observerReturnsFailingDeferred(self):
+        """
+        If the observer function returns a failing Deferred, the failure is
+        logged and processing continues.
+        """
+        dl = [defer.Deferred(), defer.Deferred()]
+        notifyD = defer.DeferredList(dl)
+
+        # use a factory function to ensure that id(observer1) != id(observer2)
+        def makeObserver():
+            def observer(notify):
+                self.notifies.append(1)
+                # callback asynchronously to make sure the function runs to
+                # completion before continuing notifyD's callback chain
+                reactor.callLater(0, dl.pop().callback, None)
+                return defer.fail(RuntimeError("boom"))
+
+            return observer
+
+        self.conn.addNotifyObserver(makeObserver())
+        self.conn.addNotifyObserver(makeObserver())
+
+        d = self.conn.runOperation("listen txpostgres_test")
+        d.addCallback(lambda _: self.sendNotify())
+        # even though the observer functions return a failing Deferred, both of
+        # them will be called
+        d.addCallback(lambda _: notifyD)
+        # both errors get logged
+        d.addCallback(lambda _: self.assertEquals(
+                len(self.flushLoggedErrors(RuntimeError)), 2))
+        return d.addCallback(lambda _: self.assertEquals(
+                self.notifies, [1, 1]))
+
+    def test_customNotifyCooperator(self):
+        """
+        Using a custom cooperator for notifies is possible.
+        """
+        cooperateD = defer.Deferred()
+
+        class FakeCooperator(object):
+
+            def cooperate(self, iterator):
+                self.result = []
+
+                for ret in iterator:
+                    defer.maybeDeferred(lambda: ret).addCallback(
+                        self.result.append)
+
+                cooperateD.callback(None)
+
+        cooperator = FakeCooperator()
+
+        def observer(notify):
+            return True
+
+        c = txpostgres.Connection(cooperator=cooperator)
+        c.addNotifyObserver(observer)
+
+        d = c.connect(user=DB_USER, password=DB_PASS,
+                      host=DB_HOST, database=DB_NAME)
+        d.addCallback(lambda _: c.runOperation("listen txpostgres_test"))
+        d.addCallback(lambda _: self.sendNotify())
+        d.addCallback(lambda _: cooperateD)
+        d.addCallback(lambda _: self.assertEquals(
+                cooperator.result, [True]))
+        return d.addCallback(lambda _: c.close())
